/*
** File: interrupt.c  
** Project: VxWorks Driver Framework
** Purpose: Interrupt hooking functions
**
** (C) Copyright Alpha Data 2010
*/

#include "df.h"
#include "dfvxworks.h"

#if defined(HCD5220)
# if (_WRS_VXWORKS_MAJOR > 6) || (_WRS_VXWORKS_MAJOR == 6 && _WRS_VXWORKS_MINOR >= 9)
#   /* vxbPciIntConnect in <hwif/vxbus/vxbPciLib.h> in VxWorks 6.9 and later */
#   include <hwif/vxbus/vxbPciLib.h>
# else
#   /* vxbPciIntConnect in <../src/hwif/h/vxbus/vxbPciBus.h> in VxWorks 6.9 and later */
#   include <../src/hwif/h/vxbus/vxbPciBus.h>
# endif
#endif

void
dfVxwIsrStub(
  DfInterruptObject* pInterruptObject)
{
  /* boolean_t bHandled; */

  dfDebugPrint(9, ("dfVxwIsrStub: entered, pInterruptObject=%p\n", pInterruptObject));

  dfAssert(pInterruptObject->system.bConnected);

#if DFVXW_USE_SPINLOCKLIB
# if defined(SPIN_LOCK_TRACE)
  spinLockIsrTake(&pInterruptObject->system.lock, __FILE__, __LINE__);
# else
  spinLockIsrTake(&pInterruptObject->system.lock);
# endif
#endif
  (void)/* bHandled = */pInterruptObject->system.pIsr(pInterruptObject, pInterruptObject->system.pContext);
#if DFVXW_USE_SPINLOCKLIB
  spinLockIsrGive(&pInterruptObject->system.lock);
#endif
  /*
  ** There's currently no means of passing any information back to the VxWorks
  ** kernel about whether or not the interrupt was handled by us, so we just
  ** discard the value of 'bHandled'.
  */
}

#if defined(ADB3_VXBUS)

/* VxBus version */

#if defined(HCD5220)
# if defined(ADB3_VXBUS3)
#   define pciConfigRead8(pDev, addr, pVal8, pFlags) pDev->pAccess->busCfgRead(pDev, addr, sizeof(*pVal8), (char*)pVal8, pFlags)
# elif defined(ADB3_VXBUS4)
#   define pciConfigRead8(pDev, addr, pVal8, pFlags) vxbPciDevCfgRead(pDev, addr, sizeof(*pVal8), pVal8)
# else
#   error Neither ADB3_VXBUS3 nor ADB3_VXBUS4 defined!
# endif
#endif

boolean_t
dfInterruptConnect(
  DfDeviceObject* pDevObj,
  DfInterruptObject* pInterruptObject,
  DfBusResources* pBusResources,
  DfBusType busType,
  unsigned int interruptIndex,
  DfIsr* pIsr,
  void* pContext)
{
  VXB_DEVICE_ID pVxbDev = pDevObj->system.pVxbDev;

  dfAssert(busType == DfBusPci);
  dfAssert(!pInterruptObject->system.bConnected);
  dfAssert(interruptIndex < pBusResources->variant.pci.numIrq);

#if defined(HCD5220)
  {
    int iNum = 0;
    UINT32 f = 0;
    uint8_t irqLine;
    STATUS status;

    pciConfigRead8(pVxbDev, PCI_CFG_DEV_INT_LINE, &irqLine, &f);
    iNum = (int)irqLine - 1;
    dfDebugPrint(2, ("dfInterruptConnect(HCD5520): pVxbDev=%p isr=%p pInterruptObject=%p iNum=%d\n",
      pVxbDev, dfVxwIsrStub, pInterruptObject, iNum));
#if DFVXW_HAS_VX_USR_ARG_T
    status = vxbPciIntConnect(pVxbDev, INUM_TO_IVEC(iNum), dfVxwIsrStub, (_Vx_usr_arg_t)pInterruptObject);
#else
    status = vxbPciIntConnect(pVxbDev, INUM_TO_IVEC(iNum), dfVxwIsrStub, (int)pInterruptObject);
#endif
    if (OK == status) {
      status = intEnable(iNum);
      if (OK != status) {
        dfDebugPrint(0, ("*** dfInterruptConnect(HCD5520): intEnable failed\n"));
        /* There seems to be no vxbPciIntDisconnect or vxbPciIntDisconnect2 */
#if DFVXW_HAS_VX_USR_ARG_T
        /* vxbPciIntDisconnect2(pVxbDev, INUM_TO_IVEC(iNum), isr, (_Vx_usr_arg_t)pIsrArg); */
#else
        /* vxbPciIntDisconnect2(pVxbDev, INUM_TO_IVEC(iNum), isr, (int)pIsrArg); */
#endif
      }
    } else {
      dfDebugPrint(0, ("*** dfInterruptConnect(HCD5520): vxbPciIntConnect failed\n"));
    }
  }
#else
  dfDebugPrint(2, ("dfInterruptConnect(VxBus generic): pVxbDev=%p interruptIndex=%u isr=%p pInterruptObject=%p\n",
    pVxbDev, interruptIndex, dfVxwIsrStub, pInterruptObject));
  if (OK != vxbIntConnect(pVxbDev, (int)interruptIndex, dfVxwIsrStub, pInterruptObject)) {
    dfDebugPrint(0, ("*** dfInterruptConnect(VxBus generic): failed to connect interrupt, pVxbDev=%p index=%u\n", pVxbDev, interruptIndex));
    return FALSE;
  }
  if (OK != vxbIntEnable(pDevObj->system.pVxbDev, (int)interruptIndex, dfVxwIsrStub, pInterruptObject)) {
    dfDebugPrint(0, ("*** dfInterruptConnect(VxBus generic): failed to enable interrupt, pVxbDev=%p interruptIndex=%u\n", pVxbDev, interruptIndex));
    if (OK != vxbIntDisconnect(pVxbDev, (int)interruptIndex, dfVxwIsrStub, pInterruptObject)) {
      dfDebugPrint(0, ("*** dfInterruptConnect(VxBus generic): failed to disconnect interrupt, pVxbDev=%p interruptIndex=%u\n", pVxbDev, interruptIndex));
    }
    return FALSE;
  }
#endif
#if DFVXW_USE_SPINLOCKLIB
  spinLockIsrInit(&pInterruptObject->system.lock, 0);
#endif
  pInterruptObject->system.pIsr = pIsr;
  pInterruptObject->system.pContext = pContext;
  pInterruptObject->system.bConnected = TRUE;
  pInterruptObject->system.vxbus.index = interruptIndex;
#if DFVXW_USE_ISRDEFERLIB
  pInterruptObject->system.isrDeferQueueId = isrDeferQueueGet(pVxbDev, 0, 0, 0);
#endif
  return TRUE;
}

void
dfInterruptDisconnect(
  DfDeviceObject* pDevObj,
  DfInterruptObject* pInterruptObject)
{
  VXB_DEVICE_ID pVxbDev = pDevObj->system.pVxbDev;
  unsigned int interruptIndex = pInterruptObject->system.vxbus.index;

  dfAssert(pInterruptObject->system.bConnected);

  dfDebugPrint(2, ("dfInterruptConnect: pInterruptObject=%p index=%u\n", pInterruptObject, interruptIndex));

#if defined(HCD5220)
  {
    int iNum = 0;
    UINT32 f = 0;
    uint8_t irqLine;

    pciConfigRead8(pVxbDev, PCI_CFG_DEV_INT_LINE, &irqLine, &f);
    iNum = (int)irqLine - 1;
    dfDebugPrint(2, ("dfInterruptConnect(HCD5220): pVxbDev=%p isr=%p pInterruptObject=%p iNum=%d\n",
      pVxbDev, dfVxwIsrStub, pInterruptObject, iNum));
    intDisable(iNum);
    /* There seems to be no vxbPciIntDisconnect or vxbPciIntDisconnect2 */
#if DFVXW_HAS_VX_USR_ARG_T
    /* vxbPciIntDisconnect2(pVxbDev, INUM_TO_IVEC(iNum), dfVxwIsrStub, (_Vx_usr_arg_t)pInterruptObject); */
#else
    /* vxbPciIntDisconnect2(pVxbDev, INUM_TO_IVEC(iNum), dfVxwIsrStub, (int)pInterruptObject); */
#endif
  }
#else
  if (OK != vxbIntDisable(pVxbDev, (int)interruptIndex, dfVxwIsrStub, pInterruptObject)) {
    dfDebugPrint(0, ("*** dfInterruptDisconnect: failed to disable interrupt, pVxbDev=%p index=%u\n",
      pDevObj->system.pVxbDev, pInterruptObject->system.vxbus.index));
  }
  if (OK != vxbIntDisconnect(pVxbDev, (int)interruptIndex, dfVxwIsrStub, pInterruptObject)) {
    dfDebugPrint(0, ("*** dfInterruptDisconnect: failed to disconnect interrupt, pVxbDev=%p index=%u\n",
      pDevObj->system.pVxbDev, pInterruptObject->system.vxbus.index));
  }
#endif
  pInterruptObject->system.bConnected = FALSE;
}

#else

boolean_t
dfInterruptConnect(
  DfDeviceObject* pDevObj,
  DfInterruptObject* pInterruptObject,
  DfBusResources* pBusResources,
  DfBusType busType,
  unsigned int interruptIndex,
  DfIsr* pIsr,
  void* pContext)
{
  dfAssert(!pInterruptObject->system.bConnected);
  dfAssert(interruptIndex < pBusResources->variant.pci.numIrq);

#if DFVXW_USE_SPINLOCKLIB
  spinLockIsrInit(&pInterruptObject->system.lock, 0);
#endif
  pInterruptObject->system.pIsr = pIsr;
  pInterruptObject->system.pContext = pContext;

  if (!dfVxwLegacyInterruptConnect(pDevObj, pInterruptObject, pBusResources, busType, interruptIndex, pIsr, pContext)) {
    return FALSE;
  }
  pInterruptObject->system.bConnected = TRUE;
#if DFVXW_USE_ISRDEFERLIB
  /* Not sure if passing NULL is legal... but DFVXW_USE_ISRDEFERLIB is 0 anyway */
  pInterruptObject->system.isrDeferQueueId = isrDeferQueueGet(NULL, 0, 0, 0);
#endif
  return TRUE;
}

void
dfInterruptDisconnect(
  DfDeviceObject* pDevObj,
  DfInterruptObject* pInterruptObject)
{
  dfAssert(pInterruptObject->system.bConnected);

  dfVxwLegacyInterruptDisconnect(pDevObj, pInterruptObject);

  pInterruptObject->system.bConnected = FALSE;
}

#endif

boolean_t
dfInterruptUsingMsi(
  DfDeviceObject* pDevObj,
  DfInterruptObject* pInterruptObject)
{
  dfAssert(pInterruptObject->system.bConnected);
  return FALSE;
}
